
/* interface */
#ifndef THREAD_WIN32_H_INCLUDED
#define THREAD_WIN32_H_INCLUDED

#include <windows.h>

typedef HANDLE yarv_thread_id_t;
typedef HANDLE yarv_thread_lock_t;

int native_mutex_lock(yarv_thread_lock_t *);
int native_mutex_unlock(yarv_thread_lock_t *);
void native_mutex_initialize(yarv_thread_lock_t *);

#endif /* THREAD_WIN32_H_INCLUDED */






/**********************************************************************/

#ifdef THREAD_SYSTEM_DEPENDENT_IMPLEMENTATION

#include <process.h>

#define WIN32_WAIT_TIMEOUT 10 /* 10 ms */
#undef Sleep

static HANDLE interrupted_event;

static void
Init_native_thread(){
  interrupted_event
    = CreateEvent(NULL, FALSE, FALSE, NULL);
  GET_THREAD()->thread_id = GetCurrentThread();
}

static int system_working = 1;

static int
kill_func(st_data_t key, st_data_t val, void *dummy){
  HANDLE thid = (HANDLE)val;
  if(GetCurrentThread() != thid){
    TerminateThread(thid, 0);
  }
  return ST_CONTINUE;
}

void
native_thread_cleanup(yarv_vm_t *vm){
  system_working = 0;
  st_foreach(vm->living_threads, kill_func, 0);
}

static void
w32_show_error_message()
{
  LPVOID lpMsgBuf;
  FormatMessage(
    FORMAT_MESSAGE_ALLOCATE_BUFFER |
    FORMAT_MESSAGE_FROM_SYSTEM |
    FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    (LPTSTR) &lpMsgBuf,
    0,
    NULL);
  
  MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
  exit(1);
}

static int
w32_wait_event(HANDLE event, DWORD timeout)
{
  HANDLE events[2];
  int count = 0;
  DWORD ret;

  if(event){
    events[count++] = event;
  }
  events[count++] = interrupted_event;

  ret = WaitForMultipleObjectsEx(count, events, FALSE, timeout, TRUE);
  if(ret == -1){
    w32_show_error_message();
  }
  if (ret == WAIT_OBJECT_0 + count - 1) {
    errno = EINTR;
  }
  return ret;
}

int
native_mutex_lock(yarv_thread_lock_t *lock){
  DWORD result;
  
  while(1){
    result = w32_wait_event(*lock, INFINITE);
    switch(result){
    case WAIT_OBJECT_0:
      /* get mutex object */
      thread_debug("%p acquire mutex: %p\n", GetCurrentThreadId(), *lock);
      return 0;
    case WAIT_TIMEOUT:
      /* TODO: check interrupt */
      thread_debug("%p timeout mutex: %p\n", GetCurrentThreadId(), *lock);
      break;
    case WAIT_ABANDONED:
      rb_bug("win32_mutex_lock: WAIT_ABANDONED");
      break;
    default:
      rb_bug("win32_mutex_lock: unknown result (%d)", result);
      break;
    }
  }
  return 0;
}

int
native_mutex_unlock(yarv_thread_lock_t *lock){
  thread_debug("%p release mutex: %p\n",  GetCurrentThreadId(), *lock);
  return ReleaseMutex(*lock);
}

void
native_mutex_initialize(yarv_thread_lock_t *lock){
  *lock = CreateMutex(NULL, FALSE, NULL);
  thread_debug("%p initialize mutex: %p\n",  GetCurrentThreadId(), *lock);
}

static int
native_thread_join(yarv_thread_id_t thid, void *ptr){
  DWORD result;

  while(1){
    result = w32_wait_event(thid, INFINITE);
    switch(result){
    case WAIT_OBJECT_0:
      /* join */
      if(ptr){
        GetExitCodeThread(thid, ptr);
      }
      thread_debug("%p join to %p\n", GetCurrentThreadId(), thid);
      return 0;
    case WAIT_TIMEOUT:
      /* TODO: check interrupt */
      thread_debug("%p join timeout (to %p)\n", GetCurrentThreadId(), thid);
      break;
    default:
      rb_bug("native_thread_join: unknown result (%d)", result);
      break;
    }
  }
  return 1;
}


NOINLINE(static int thread_start_func_2(yarv_thread_t *th, VALUE *stack_start));
void static thread_cleanup_func(void *th_ptr);

static int _stdcall
thread_start_func_1(void *th_ptr)
{
  yarv_thread_t *th = th_ptr;
  VALUE stack_start;
  /* run */
  thread_debug("%p (th: %p) thread created\n", GetCurrentThreadId(), th);
  thread_start_func_2(th, &stack_start);
  thread_cleanup_func(th);
  thread_debug("%p (th: %p) thread deleted\n", GetCurrentThreadId(), th);
  return 0;
}

static void make_timer_thread();

static int
native_thread_crteate(yarv_thread_t *th)
{
  size_t stack_size = 4 * 1024 - sizeof(int); /* 4KB */
  static int init = 0;

  if(!init){
    make_timer_thread();
  }
  
  if((th->thread_id = (HANDLE)_beginthreadex(0, /* security */
                                             stack_size,
                                             thread_start_func_1,
                                             th,
                                             0, /* init flag */
                                             0  /* handle */)) == 0){
    rb_raise(rb_eThreadError, "can't create Thread (%d)", errno);
  }
  thread_debug("%p create: %p (th: %p), stack size: %d\n",  GetCurrentThreadId(), th->thread_id, th, stack_size);
  return 0;
}

static HANDLE timer_thread_handle = 0;

static int _stdcall
timer_thread_func(void *dummy){
  thread_debug("timer_thread: %p\n", GetCurrentThreadId());
  while(system_working){
    Sleep(WIN32_WAIT_TIMEOUT);
    GET_VM()->interrupt_flag = 1;
  }
  return 0;
}

static void
make_timer_thread(){
  if(timer_thread_handle == 0){
    timer_thread_handle =
      (HANDLE)_beginthreadex(0,
                             1024,
                             timer_thread_func,
                             0, 0, 0);
  }
}

#endif /* THREAD_SYSTEM_DEPENDENT_IMPLEMENTATION */
